@@ -46,7 +46,7 @@ module NumRu
     def eof(gphys, *args)
 
       unless defined?(@@EOF_engin)
-        raise "eigher ssl2 or gsl is required"
+        raise "SSL2 (Ruby-SSL2) or LAPACK (Ruby-LAPACK) or GSL (Ruby/GSL) must have been installed. (SSL2 or LAPACK is recommended for large computation)"
       end
 
       if Hash === args[-1]
