@@ -7,54 +7,59 @@
 require "narray_miss"
 require "date"
 require "time"
+require "numru/gphys_ext"
+
+## /--->
+## backport of String#ord and String#getbyte (ruby 1.9 feature)
+## to Ruby 1.8 or earlier
+=begin
+if ! 'a'.respond_to?(:ord)
+  class String
+    def ord
+      self[0]
+    end
+  end
+end
+if ! 'a'.respond_to?(:getbyte)
+  class String
+    def getbyte(i)
+      self[i]
+    end
+  end
+end
+=end
+## <---/
 
 module NumRu
   module GribUtils
     require "numru/gphys/grib_params"
     private
+=begin
     def str2uint1(str)
       return nil if str.length==0
-      return str[0]
+      return str.ord
     end
     def str2uint2(str)
       return nil if str.length==0
-      return (str[0]<<8)+str[1]
+      return (str.ord<<8)+str.getbyte(1)
     end
     def str2uint3(str)
       return nil if str.length==0
-      return (str[0]<<16)+(str[1]<<8)+str[2]
-    end
-=begin
-    def str2uint(str)
-      return nil until str
-      return str.unpack("H*")[0].hex
+      return (str.ord<<16)+(str.getbyte(1)<<8)+str.getbyte(2)
     end
 =end
     def str2int1(str)
-      n = str2uint1(str)
+      n = str.to_uint1
       return n && n > 127 ? 128-n : n
     end
     def str2int2(str)
-      n = str2uint2(str)
+      n = str.to_uint2
       return n && n > 32767 ? 32768-n : n
     end
     def str2int3(str)
-      n = str2uint3(str)
+      n = str.to_uint3
       return n && n > 8388607 ? 8388608-n : n
     end
-=begin
-    def str2int(str)
-      return nil if str.length==0
-      ary = str.unpack("H*")
-      top = ary[0]
-      fact = 1
-      if top[0,1]>"7"
-        ary[0][0] = (top[0,1].hex-8).to_s
-        fact = -1
-      end
-      return ary.join("").hex*fact
-    end
-=end
     def uint2str(i,len)
       str = [sprintf("%0#{len*2}x",i)].pack("H*")
       str.length==len || raise("str length is not valid")
@@ -71,10 +76,10 @@ module NumRu
     end
     def float_value( str )
       str.length==4 || raise("str length must be 4")
-      sa = str2int1( str[0,1] )
+      sa = str2int1(str)
       s = (sa<0) ? -1 : 1
       a = sa.abs
-      b = str2uint3( str[1,3] )
+      b = str.to_uint3(1)
       return s*(2.0)**(-24.0)*b*(16.0)**(a-64.0)
     end
     def get_bits( str, nstep, nbits, sbu, ebu, type )
@@ -176,7 +181,7 @@ module NumRu
       return d*24+h+m/60
     end
 
-  end
+  end # GribUtils
 
   class Grib
     private
@@ -195,12 +200,13 @@ module NumRu
           obj.is.update_total_length
           return obj
         end
-        def parse(file)
+        def parse(file, mbio)
           obj = GribSegment.new
           obj.file = file
+          obj.mbio = mbio
           obj.pos = file.pos
 
-          is = GribIS.new(obj,file.read(8))
+          is = GribIS.new(obj,ss = file.read(8))
           unless is.version == 1
             warn "This Grib version (#{is.version}) has not supported yet"
             return nil
@@ -210,18 +216,18 @@ module NumRu
           end
           obj.is = is
 
-          pds = GribPDS.new(obj,file.read(str2uint3(file.read(3))-3))
+          pds = GribPDS.new(obj,file.read(file.read(3).to_uint3-3))
           obj.pds = pds
 
-          str = pds.gds? ? file.read(str2uint3(file.read(3))-3) : nil
+          str = pds.gds? ? file.read(file.read(3).to_uint3-3) : nil
           gds = GribGDS.new(obj,str)
           obj.gds = gds
 
-          str = pds.bms? ? file.read(str2uint3(file.read(3))-3) : nil
+          str = pds.bms? ? file.read(file.read(3).to_uint3-3) : nil
           bms = GribBMS.new(obj,str)
           obj.bms = bms
 
-          len = str2uint3(file.read(3))
+          len = file.read(3).to_uint3
           bds = GribBDS.new(obj,file.pos,len)
           obj.bds = bds
           file.seek(len-3,IO::SEEK_CUR)
@@ -233,13 +239,13 @@ module NumRu
           obj.es = es
 
           if is.total_length!=is.length+pds.length+gds.length+bms.length+bds.length+es.length
-            raise "total length is not equal to the sum of each section"
+            raise "total length is not equal to the sum of each section\nfile may be broken"
           end
 
           return obj
         end
       end
-      attr_accessor :file, :pos
+      attr_accessor :file, :pos, :mbio
       attr_accessor :is, :pds, :gds, :bms, :bds, :es
       def set_xy(x,y)
         pds.set_gds(true)
@@ -360,7 +366,7 @@ module NumRu
         return @is[0..3]=="GRIB"
       end
       def total_length
-        return str2uint3( @is[4,3] )
+        return @is.to_uint3(4)
       end
       def update_total_length
         len = self.length+@sgm.pds.length+@sgm.gds.length+@sgm.bms.length+@sgm.bds.length+@sgm.es.length
@@ -368,7 +374,7 @@ module NumRu
         return len
       end
       def version
-        return str2uint1( @is[7,1] )
+        return @is.to_uint1(7)
       end
       def set_version(ver)
         @is[7..7] = uint2str(ver,1)
@@ -383,7 +389,7 @@ module NumRu
       include GribUtils
       private
       def param
-        nid = str2uint1(@pds[5,1])
+        nid = @pds.to_uint1(5)
         params = nil
         case version
         when 0
@@ -414,10 +420,10 @@ module NumRu
             if para = NAMES_UNITS[para_id]
               return para
             else
-              warn "parameter ID #{nid} in version #{version} has not defined yet"
+              warn "parameter ID #{nid} in table version #{version} has not defined yet"
             end
           else
-            warn "parameter ID #{nid} in version #{version} has not defined yet"
+            warn "parameter ID #{nid} in table version #{version} has not defined yet"
           end
         else
           warn "parameter table version #{version}, center id #{cid}, version #{version}, parameter id #{nid} has not defined yet"
@@ -425,7 +431,7 @@ module NumRu
         return ["unknown (#{nid})", nid.to_s, "-", nil]
       end
       def cid
-        return str2uint1(@pds[1,1])
+        return @pds.to_uint1(1)
       end
       def initialize(sgm,str=nil)
         @sgm = sgm
@@ -436,7 +442,7 @@ module NumRu
         @pds.length+3
       end
       def version
-        return str2uint1( @pds[0,1] )
+        return @pds.to_uint1
       end
       def set_version(ver)
         @pds[0..0] = uint2str(ver,1)
@@ -451,7 +457,7 @@ module NumRu
       end
       alias :center_id= :set_center_id
       def pid
-        return str2uint1( @pds[2,1] )
+        return @pds.to_uint1(2)
       end
       def set_pid(id)
         @pds[2..2] = uint2str(id,1)
@@ -459,7 +465,7 @@ module NumRu
       end
       alias :pid= :set_pid
       def gid
-        return str2uint1( @pds[3,1] )
+        return @pds.to_uint1(3)
       end
       def set_gid(id)
         @pds[3..3] = uint2str(id,1)
@@ -467,7 +473,7 @@ module NumRu
       end
       alias :gid= :set_gid
       def gds?
-        flag = str2uint1( @pds[4,1] )
+        flag = @pds.to_uint1(4)
         return (flag>>7)&1==1
       end
       def set_gds(gds)
@@ -486,7 +492,7 @@ module NumRu
       end
       alias :gds= :set_gds
       def bms?
-        flag = str2uint1( @pds[4,1] )
+        flag = @pds.to_uint1(4)
         return (flag>>6)&1==1
       end
       def set_bms(bms)
@@ -515,7 +521,7 @@ module NumRu
       end
       alias :name_id= :set_name_id
       def sname
-        zid = str2uint1( @pds[6,1] )
+        zid = @pds.to_uint1(6)
         sn = param[1]
         zt = Z_TYPES[zid]
         zn = zt ? zt[1] : zid.to_s
@@ -525,7 +531,7 @@ module NumRu
         return param[2]
       end
       def z_type
-        zid = str2uint1( @pds[6,1] )
+        zid = @pds.to_uint1(6)
         zt = Z_TYPES[zid]
         if zt
           return zt[0]
@@ -540,7 +546,7 @@ module NumRu
       end
       alias :z_id= :set_z_id
       def z_sname
-        zid = str2uint1( @pds[6,1] )
+        zid = @pds.to_uint1(6)
         if zn = Z_TYPES[zid]
           return zn[1]
         else
@@ -549,7 +555,8 @@ module NumRu
         end
       end
       def z_value
-        type = str2uint1( @pds[6,1] )
+        return @z_value if @z_value
+        type = @pds.to_uint1(6)
         str = @pds[7,2]
         ary = Z_LEVELS[type]
         if ary.nil?
@@ -558,15 +565,16 @@ module NumRu
           ary = ary.dup
         end
         if ary.length==1
-          ary[0] = ary[0].dup.update( {"value"=> str2uint2(str)} )
+          ary[0] = ary[0].dup.update( {"value"=> str.to_uint2} )
         elsif ary.length==2
-          ary[0] = ary[0].dup.update( {"value"=> str2uint1(str[0,1])} )
-          ary[1] = ary[1].dup.update( {"value"=> str2uint1(str[1,1])} )
+          ary[0] = ary[0].dup.update( {"value"=> str.to_uint1} )
+          ary[1] = ary[1].dup.update( {"value"=> str.to_uint1(1)} )
         end
+        @z_value = ary
         return ary
       end
       def set_z_value(val)
-        type = str2uint1( @pds[6,1] )
+        type = @pds.to_uint1(6)
         ary = Z_LEVELS[type]
         val.length==ary.length || raise("length of val is not collect")
         if val.length==1
@@ -581,7 +589,7 @@ module NumRu
       end
       alias :z_value= :set_z_value
       def initial
-        return [str2uint1( @pds[21,1] ), str2uint1( @pds[9,1] ), str2uint1( @pds[10,1] ), str2uint1( @pds[11,1] ), str2uint1( @pds[12,1] ), str2uint1( @pds[13,1] )]
+        return [@pds.to_uint1(21), @pds.to_uint1(9), @pds.to_uint1(10), @pds.to_uint1(11),  @pds.to_uint1(12), @pds.to_uint1(13)]
       end
       def set_initial(year,month,day,hour,min)
         cent = year/100+1
@@ -596,14 +604,14 @@ module NumRu
       end
       alias :initial= :set_initial
       def date
-        d = Date.new((str2uint1( @pds[21,1] )-1)*100+str2uint1( @pds[9,1] ),
-                     str2uint1( @pds[10,1] ),
-                     str2uint1( @pds[11,1] ) )
-        h = [str2uint1( @pds[12,1] ),str2uint1( @pds[13,1] )]
+        d = Date.new((@pds.to_uint1(21)-1)*100 + @pds.to_uint1(9),
+                     @pds.to_uint1(10), @pds.to_uint1(11)
+                     )
+        h = [@pds.to_uint1(12), @pds.to_uint1(13)]
         return [d,h]
       end
       def time_unit
-        return TIME_UNITS[ str2uint1( @pds[14,1] ) ]
+        return TIME_UNITS[ @pds.to_uint1(14) ]
       end
       def set_time_unit_id(id)
         @pds[14,1] = uint2str(id,1)
@@ -611,7 +619,7 @@ module NumRu
       end
       alias :time_unit_id= :set_time_unit_id
       def p1
-        return str2uint1( @pds[15,1] )
+        return @pds.to_uint1(15)
       end
       def set_p1(i)
         @pds[15,1] = uint2str(i,1)
@@ -619,7 +627,7 @@ module NumRu
       end
       alias :p1= :set_p1
       def p2
-        return str2uint1( @pds[16,1] )
+        return @pds.to_uint1(16)
       end
       def set_p2(i)
         @pds[16,1] = uint2str(i,1)
@@ -627,7 +635,7 @@ module NumRu
       end
       alias :p2= :set_p2
       def time_range
-        return str2uint1( @pds[17,1] )
+        return @pds.to_uint1(17)
       end
       def set_time_range_id(id)
         @pds[17,1] = uint2str(id,1)
@@ -635,7 +643,7 @@ module NumRu
       end
       alias :time_range_id= :set_time_range_id
       def ave
-        return str2uint2( @pds[18,2] )
+        return @pds.to_uint2(18)
       end
       def set_ave(val)
         @pds[18,2] = uint2str(val,2)
@@ -643,7 +651,7 @@ module NumRu
       end
       alias :ave= :set_ave
       def miss
-        return str2uint1( @pds[20,1] )
+        return @pds.to_uint1(20)
       end
       def set_miss(val)
         @pds[20,1] = uint2str(val,1)
@@ -651,7 +659,7 @@ module NumRu
       end
       alias :miss= :set_miss
       def sub_center
-        return str2uint1( @pds[22,1] )
+        return @pds.to_uint1(22)
       end
       def set_sub_center_id(id)
         @pds[22,1] = uint2str(id,1)
@@ -674,29 +682,29 @@ module NumRu
       include GribUtils
       private
       def gtype
-        return str2uint1( @gds[2,1] )
+        return @gds.to_uint1(2)
       end
       def get_x_y(grid)
       case gtype
         when 0,4
           #lon_lat (+GussLat)
-          flag = str2uint1( grid[10,1] )
+          flag = grid.to_uint1(10)
           flag_inc = (flag>>7)&1
-          flag = str2uint1( grid[21,1] )
+          flag = grid.to_uint1(21)
           flag_scani = (flag>>7)&1
-          flag_scanj = (flag>>6)&1
+          flag_scanj = (flag>>6)^1
           flag_ij = (flag>>5)&1
 
           lon = Hash.new
           lon["short_name"] = "lon"
           lon["long_name"] = "longitude"
           lon["units"] = "degrees_east"
-          nlon = str2uint2( grid[0,2] )
-          nlon=="ffff".hex && raise("not defined yet")
+          nlon = grid.to_uint2
+          nlon==0xffff && raise("not defined yet")
           lo1 = str2int3( grid[7,3] )
           lo2 = str2int3( grid[14,3] )
-          dlon = str2uint2( grid[17,2] )
-          if (!flag_inc)&&dlon==-"7fff".hex
+          dlon = grid.to_uint2(17)
+          if (!flag_inc)&&dlon==-0x7fff
             dlon = (lo2-lo1)/(nlon-1)
           else
             if lo1>lo2
@@ -710,7 +718,7 @@ module NumRu
           dlon2 = (lo2-lo1).to_f/(nlon-1)
           dlon2.round==dlon.round || warn("Warning: dlon is not same: #{dlon2/1000} != #{dlon.to_f/1000}")
           vlon = NArray.sfloat(nlon).indgen*dlon2+lo1
-          if flag_scani==1
+          if flag_scani==1 && dlon>0 # THIS PART MIGHT BE MEANINGLESS (OR WRONG)
             vlon = vlon[-1..0]
           end
           lon["value"] = vlon/1000
@@ -720,13 +728,13 @@ module NumRu
           lat["short_name"] = "lat"
           lat["long_name"] = "latitude"
           lat["units"] = "degrees_north"
-          nlat = str2uint2( grid[2,2] )
-          nlat=="ffff".hex && raise("not defined yet")
+          nlat = grid.to_uint2(2)
+          nlat==0xffff && raise("not defined yet")
           la1 = str2int3( grid[4,3] )
           la2 = str2int3( grid[11,3] )
-          dlat = str2uint2( grid[19,2] )
+          dlat = grid.to_uint2(19)
           if gtype==0
-            if (!flag_inc)&&dlat==-"7fff".hex
+            if (!flag_inc)&&dlat==-0x7fff
               dlat = (la2-la1)/(nlat-1)
             else
               dlat = -dlat if la1>la2
@@ -755,7 +763,7 @@ module NumRu
             lat["type"] = "Gaussian latitude"
             lat["number_of_latitude_circles"] = NArray[dlat*2]
           end
-          if flag_scanj==1
+          if flag_scanj==1 && dlat>0 # THIS PART MIGHT BE MEANINGLESS (OR WRONG)
             vlat = vlat[-1..0]
           end
           lat["value"] = vlat
@@ -862,15 +870,15 @@ module NumRu
         end
       end
       def nv
-        return str2uint1( @gds[0,1] )
+        return @gds.to_uint1
       end
       def pv
-        pv = str2uint1( @gds[1,1] )
+        pv = @gds.to_uint1(1)
         return nil if pv==255 || nv==0
         return pv
       end
       def pl
-        pl = str2uint1( @gds[1,1] )
+        pl = @gds.to_uint1(1)
         return nil if pl==255
         return nv*4+pl
       end
@@ -909,7 +917,7 @@ module NumRu
       end
       def list_pv
         return nil unless pv
-        list = NArray.float(nv)
+        list = NArray.sfloat(nv)
         nv.times{|n|
           list[n] = float_value( @gds[pv-4+n*4...pv-4+(n+1)*4] )
         }
@@ -979,7 +987,7 @@ module NumRu
         end
       end
       def map_number
-        nmap = str2uint2( @bms[1,2] )
+        nmap = @bms.to_uint2(1)
         nmap = false if nmap==0
         return nmap
       end
@@ -994,7 +1002,7 @@ module NumRu
       end
       def map
         return nil unless has_map?
-        return get_bits( @bms[3..-1], 1,1, 0,str2uint1(@bms[0,1]), NArray::BYTE )
+        return get_bits( @bms[3..-1], 1,1, 0,@bms.to_uint1, NArray::BYTE )
       end
       def set_map(mask)
         map,n = data2str(mask,1,1)
@@ -1013,14 +1021,14 @@ module NumRu
       def oct4
         return @oct4 if @oct4
         @sgm.file.seek(@pos, IO::SEEK_SET)
-        @oct4 = str2uint1( @sgm.file.read(1) ) 
+        @oct4 = @sgm.file.read(1).to_uint1
         return @oct4
       end
       def oct14
         return nil if flag4==0
         return @oct14 if @oct14
         @sgm.file.seek(@pos+10, IO::SEEK_SET)
-        @oct14 = str2uint1( @sgm.file.read(1) )
+        @oct14 = @sgm.file.read(1).to_uint1
         return @oct14
       end
       def flag1
@@ -1056,15 +1064,15 @@ module NumRu
         end
       end
       def get_value(file,pos,sb,eb,nbits,nstep)
-        sbu = (sb%8)/nstep
+        sbu = (sb&7)/nstep
         sb = sb/8
-        ebu = eb%8
-        ebu = (8-eb%8)/nstep unless ebu==0
+        ebu = eb&7
+        ebu = (8-(eb&7))/nstep unless ebu==0
         eb = (eb+7)/8
-        return NArray.float(0) if (eb-sb)==0
+        return NArray.sfloat(0) if eb==sb
         file.seek(pos+sb,IO::SEEK_SET)
         str = file.read(eb-sb)
-        return get_bits(str,nstep,nbits,sbu,ebu,NArray::FLOAT)
+        return get_bits(str,nstep,nbits,sbu,ebu,NArray::SFLOAT)
       end
 
       public
@@ -1079,7 +1087,7 @@ module NumRu
         return flag1==0
       end
       def set_grid(l)
-        @oct4 = oct4&"7f".hex
+        @oct4 = oct4&0x7f
         @oct4 += 128 unless l
         @bds[0,1] = uint2str(@oct4,1)
         return l
@@ -1088,7 +1096,7 @@ module NumRu
         return flag1==1
       end
       def set_spectrum(l)
-        @oct4 = oct4&"7f".hex
+        @oct4 = oct4&0x7f
         @oct4 += 128 if l
         @bds[0,1] = uint2str(@oct4,1)
         return l
@@ -1097,7 +1105,7 @@ module NumRu
         return flag2==0
       end
       def set_simple(l)
-        @oct4 = oct4&"bf".hex
+        @oct4 = oct4&0xbf
         @oct4 += 64 unless l
         @bds[0,1] = uint2str(@oct4,1)
         return l
@@ -1106,7 +1114,7 @@ module NumRu
         return flag2==1
       end
       def set_complex(l)
-        @oct4 = oct4&"bf".hex
+        @oct4 = oct4&0xbf
         @oct4 += 64 if l
         @bds[0,1] = uint2str(@oct4,1)
         return l
@@ -1115,7 +1123,7 @@ module NumRu
         return flag3==0
       end
       def set_float(l)
-        @oct4 = oct4&"df".hex
+        @oct4 = oct4&0xdf
         @oct4 += 32 unless l
         @bds[0,1] = uint2str(@oct4,1)
         return l
@@ -1124,7 +1132,7 @@ module NumRu
         return flag3==1
       end
       def set_integer(l)
-        @oct4 = oct4&"df".hex
+        @oct4 = oct4&0xdf
         @oct4 += 32 if l
         @bds[0,1] = uint2str(@oct4,1)
         return l
@@ -1133,9 +1141,9 @@ module NumRu
         return flag6 && flag6==0
       end
       def set_single(l)
-        @oct14 = oct14&"7f".hex
+        @oct14 = oct14&0x7f
         @oct14 += 128 unless l
-        @oct4 = (oct4&"ef".hex)+16
+        @oct4 = (oct4&0xef)+16
         @bds[0,1] = uint2str(@oct4,1)
         @bds[10,1] = uint2str(@oct14,1)
         return l
@@ -1144,9 +1152,9 @@ module NumRu
         return flag6 && flag6==1
       end
       def set_matrix(l)
-        @oct14 = oct14&"7f".hex
+        @oct14 = oct14&0x7f
         @oct14 += 128 if l
-        @oct4 = (oct4&"ef".hex)+16
+        @oct4 = (oct4&0xef)+16
         @bds[0,1] = uint2str(@oct4,1)
         @bds[10,1] = uint2str(@oct14,1)
         return l
@@ -1155,9 +1163,9 @@ module NumRu
         return flag7 && flag7==1
       end
       def set_bit_map(l)
-        @oct14 = oct14&"bf".hex
+        @oct14 = oct14&0xbf
         @oct14 += 64 if l
-        @oct4 = (oct4&"ef".hex)+16
+        @oct4 = (oct4&0xef)+16
         @bds[0,1] = uint2str(@oct4,1)
         @bds[10,1] = uint2str(@oct14,1)
         return l
@@ -1166,9 +1174,9 @@ module NumRu
         return flag8 && flag8==0
       end
       def set_constant(l)
-        @oct14 = oct14&"df".hex
+        @oct14 = oct14&0xdf
         @oct14 += 32 unless l
-        @oct4 = (oct4&"ef".hex)+16
+        @oct4 = (oct4&0xef)+16
         @bds[0,1] = uint2str(@oct4,1)
         @bds[10,1] = uint2str(@oct14,1)
         return l
@@ -1188,7 +1196,7 @@ module NumRu
       def value(shape,*arg)
         nbits_unuse = oct4&15
         @sgm.file.seek(@pos+7, IO::SEEK_SET)
-        nbits_pack = str2uint1( @sgm.file.read(1) )
+        nbits_pack = @sgm.file.read(1).to_uint1
         if (nbits_pack%8) == 0
           nstep = 8
         elsif (nbits_pack%4) == 0
@@ -1224,9 +1232,12 @@ module NumRu
             j = nil
             i = nil
             if map
-              var = NArrayMiss.float(*shape2)
+              var = NArrayMiss.sfloat(*shape2)
               map.reshape!(nlon,nlat)
               mask = map[*index]
+              nd = mask.count_true
+              var[mask.where] = @sgm.mbio.read2D(@pos+8, nbits_pack, nd, 1, 0, nd-1, 1, 0, 0, 1, nil, nil, e, r).reshape!(nd)
+=begin
               index[1].each{|j|
                 jj = j-index[1].first
                 sb = j*nlon+index[0].first
@@ -1237,35 +1248,43 @@ module NumRu
                   var[mask[true,j].where,jj] = get_value(@sgm.file,@pos+8,sb,eb,nbits_pack,nstep)
                 end
               }
+=end
             else
-              var = NArray.float(*shape2)
-              index[1].each{|j|
-                sb = (j*nlon+index[0].first)*nbits_pack
-                eb = (j*nlon+index[0].end+1)*nbits_pack
-                jj = j-index[1].first
-                var[true,jj] = get_value(@sgm.file,@pos+8,sb,eb,nbits_pack,nstep)
-              }
+              var = @sgm.mbio.read2D(@pos+8, nbits_pack, nlon, nlat, index[0].first, index[0].end, 1, index[1].first, index[1].end, 1, nil, nil, e, r)
+=begin
+              sb = index[1].first*nlon*nbits_pack
+              eb = (index[1].end+1)*nlon*nbits_pack
+              var = get_value(@sgm.file,@pos+8,sb,eb,nbits_pack,nstep)
+              var.reshape!(nlon, var.length/nlon)
+              var = var[index[0],true]
+=end
             end
           else
             if map
               eb = nlen*nbits_pack
-              var = NArrayMiss.float(nlon,nlat)
-              var[map.where] = get_value(@sgm.file,@pos+8,0,eb,nbits_pack,nstep)
+              var = NArrayMiss.sfloat(nlon,nlat)
+#              var[map.where] = get_value(@sgm.file,@pos+8,0,eb,nbits_pack,nstep)
+              nd = map.count_true
+              var[map.where] = @sgm.mbio.read2D(@pos+8, nbits_pack, nd, 1, 0, nd-1, 1, 0, 0, 1, nil, nil, e, r).reshape!(nd)
             else
+=begin
               eb = nlen*nbits_pack
               var = get_value(@sgm.file,@pos+8,0,eb,nbits_pack,nstep)
               var.reshape!(nlon,nlat)
+=end
+              var = @sgm.mbio.read2D(@pos+8, nbits_pack, nlon, nlat, 0, nlon-1, 1, 0, nlat-1, 1, nil, nil, e, r)
             end
           end
-          var = var*e+r
+#          var = var*e+r
           return var
         else
           raise "not defined yet"
-          n1 = str2uint2( @file.read(2) )
+          n1 = @file.read(2).to_uint2
           @sgm.file.seek(1, IO::SEEK_CUR)
-          n2 = str2uint2( @sgm.file.read(2) )
-          p1 = str2uint2( @sgm.file.read(2) )
-          p2 = str2uint2( @sgm.file.read(2) ) # @bds[15..16]
+          str = @sgm.file.read(6)
+          n2 = str.to_uint2
+          p1 = str.to_uint2(2)
+          p2 = str.to_uint2(4) # @bds[15..16]
           #        width = str2uint( @bds[18..??] )
           #        bitmap = str2uint( @bds[??+1..n1-4] )
         end
@@ -1328,6 +1347,8 @@ module NumRu
       end
     end # end definition of class GribES
 
+
+    # class Grib
     def initialize(fname,mode="r")
       case mode
       when "r","rb"
@@ -1386,7 +1407,8 @@ module NumRu
       end
       def open(fname)
         f = Grib.new(fname,"r")
-        f.parse
+        mbio = NumRu::MultibitIO.new(fname)
+        f.parse(mbio)
         return f
       end
       def create(fname)
@@ -1400,13 +1422,13 @@ module NumRu
       @attr = nil
       @file.close
     end
-    def parse
+    def parse(mbio)
       @file.rewind
       while true
         @file.read(1)=="\000" && next
         @file.eof? && break
         @file.seek(-1, IO::SEEK_CUR)
-        sgm = GribSegment.parse(@file)
+        sgm = GribSegment.parse(@file, mbio)
         return nil unless sgm
         name = sgm.pds.sname
         if @vars.has_key?(name)
@@ -1872,7 +1894,7 @@ module NumRu
       if @obj.nil?
         nil
       elsif Array===@obj
-        NArray::FLOAT
+        NArray::SFLOAT
       else
         @obj.typecode
       end
@@ -1934,7 +1956,7 @@ module NumRu
         else
           mask = true
         end
-        value = @miss ? NArrayMiss.float(*sha) : NArray.float(*sha)
+        value = @miss ? NArrayMiss.sfloat(*sha) : NArray.sfloat(*sha)
         if rank==2
           vz = nil
          vt = nil
@@ -1960,7 +1982,7 @@ module NumRu
           if vt
             index[-1] = vt.eq(get_time(pds.date)).where[0]
           end
-          next if (mask!=true && mask[*index]==0)
+          next if (NArray===mask && mask[*index]==0)
           bds = sgm.bds
           if indices.length==0 || (indices[0]==true&&indices[1]==true)
             val = bds.value(shape[0..1])
@@ -1969,7 +1991,7 @@ module NumRu
           end
           d = pds.dfact
           val = val*10.0**(-d)
-          if mask!=true
+          if NArray === mask
             index.length==first.length || raise("invalide indices")
             index.length.times{|i|
               index[i] = index[i]-first[i]
@@ -2020,9 +2042,8 @@ if $0 == __FILE__
     p v.shape
     v.att_names.each{|an| print an, " => ", v.att(an), "\n" }
     puts "\n"
+    p v.val
   }
 
 
-
-
 end
